"""PDF Generator."""
from datetime import datetime
import logging
from pathlib import Path
from typing import Dict

import jupytext
from nbconvert import PDFExporter
import pandas as pd
import papermill as pm
from pkg_resources import resource_filename

from soam.core import Step

logger = logging.getLogger(__name__)


class PDFReport:
    """Generate PDF from IPython Notebook."""

    def __init__(self, base_path: str):
        """Merge on concat dataframes dependending on the keys.

        Parameters
        ----------
        keys:
            str or list of str labels of columns to merge on
        """
        self.base_path = Path(base_path)

    def export_notebook_to_pdf(self, nb_path: str, nb_params: Dict) -> str:  # type: ignore
        """Run notebook, convert, save and return PDF path.

        Parameters
        ----------
        nb_path : str
            Path of the Notebook or Script to Execute .ipynb / .py
        parameters : Dict
            Parameters to run the notebook with (Papermill).

        Returns
        -------
        str
            Generated PDF Path
        """
        report_file = Path(nb_path)

        if not report_file.is_file():
            raise ValueError("Notebook path does not point to a file.")

        nb_params = self._parse_params(nb_params)

        report_filename, report_extension = report_file.stem, report_file.suffix

        if report_extension != "ipynb":
            report_nb = self.base_path / f"{report_filename}.ipynb"
        else:
            report_nb = report_file

        now_timestamp = datetime.today().strftime('%Y-%m-%dT%H-%M-%S')

        run_nb = self.base_path / f"{report_filename} {now_timestamp}.ipynb"
        pdf_filename = self.base_path / f"{report_filename} {now_timestamp}.pdf"

        logger.info("Running %s and converting to %s", run_nb, pdf_filename)

        jupytext_notebook = jupytext.read(report_file)
        # write to execution nb
        jupytext.write(jupytext_notebook, report_nb)

        _ = pm.execute_notebook(str(report_nb), str(run_nb), parameters=nb_params,)

        pdfexp = PDFExporter(
            template_file=resource_filename("soam", "resources/pdf_report.tpl")
        )

        pdf_data, _ = pdfexp.from_filename(run_nb)

        with open(pdf_filename, "wb") as f:
            f.write(pdf_data)

        logger.info("Succesfully wrote: %s", str(pdf_filename))

        return str(pdf_filename)

    def _parse_params(self, nb_params):
        """Convert common object types to appropiate parameter types."""
        for key, value in nb_params.items():
            if isinstance(value, pd.DataFrame):
                nb_params[key] = value.to_csv(index=False)
        return nb_params


class PDFReportTask(Step, PDFReport):
    def __init__(self, base_path: str):
        Step.__init__(self)
        PDFReport.__init__(self, base_path)

    def run(self, nb_path: str, nb_params: Dict) -> str:  # type: ignore[override]
        return self.export_notebook_to_pdf(nb_path, nb_params)



"""
Forecast Plotter
----------------
Postprocess to plot the model forecasts.
"""
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pandas as pd
from prefect.utilities.tasks import defaults_from_attrs

from soam.constants import DAILY_TIME_GRANULARITY, DS_COL
from soam.core.step import Step
from soam.plotting.plot_utils import create_forecast_figure
from soam.utilities.utils import get_file_path

logger = logging.getLogger(__name__)


class ForecastPlotterTask(Step):
    """Plot forecasts.

    Parameters
    ----------
    path : pathlib.Path or str
        Path to which plots will be saved.
    metric_name : str
        Name of the metric to plot.
    time_granularity : str
        Time granularity used on the plot. Defaults to `DAILY_TIME_GRANULARITY`.
    plot_config: dict
        Misc configs passed to `create_forecast_figure`.
    savefig_opts: dict
        kwargs passed to `savefig`.
    """

    def __init__(
        self,
        path: Union[Path, str],
        metric_name: str,
        time_granularity: str = DAILY_TIME_GRANULARITY,
        plot_config: Optional[Dict] = None,
        savefig_opts: Optional[Dict] = None,
        **kwargs: Any,
    ):
        Step.__init__(self, **kwargs)  # type: ignore
        self.path = path
        self.metric_name = metric_name
        self.time_granularity = time_granularity
        self.plot_config = plot_config
        if savefig_opts is None:
            savefig_opts = {}
        self.savefig_opts = savefig_opts

        # Last image rendered. Used for testing.
        self.fig = None

    @defaults_from_attrs(
        'path', 'metric_name', 'time_granularity', 'plot_config', 'savefig_opts'
    )
    def run(  # type: ignore
        self,
        time_series: pd.DataFrame,
        predictions: pd.DataFrame,
        path=None,
        metric_name=None,
        time_granularity=None,
        plot_config=None,
        savefig_opts=None,
    ) -> Path:
        """
        Create and store the result plot in the constructed path.

        If the path does not exist, it will be created.

        Parameters
        ----------
        time_series
            Dataframe belonging to a time_series of data.
        predictions
            Dataframe with the result of the predictions.

        Returns
        -------
        pathlib.Path
            The path of the resulting plot
        """

        full_series = pd.concat([predictions, time_series])
        full_series[DS_COL] = pd.to_datetime(full_series[DS_COL])
        start_date = min(pd.to_datetime(time_series[DS_COL]))
        end_date = pd.to_datetime(predictions[DS_COL]).min()

        forecast_window = (pd.to_datetime(predictions[DS_COL]).max() - end_date).days

        fig = create_forecast_figure(
            full_series,
            metric_name,
            end_date,
            forecast_window,
            time_granularity=time_granularity,
            plot_config=plot_config,
        )

        path.mkdir(parents=True, exist_ok=True)
        fn = "_".join(
            ["forecast", f"{start_date:%Y%m%d%H}", f"{end_date:%Y%m%d%H}", ".png"]
        )
        plot_path = get_file_path(path, fn)
        logger.debug(f"Saving forecast figure to {plot_path}...")
        fig.savefig(plot_path, bbox_inches="tight", **savefig_opts)
        self.fig = fig
        return plot_path



"""
MergeConcat
-----------
A class to merge or concat dataframes
"""

from typing import List, Union

import pandas as pd
from pandas.core.common import maybe_make_list

from soam.core import Step


class MergeConcat(Step):
    def __init__(
        self, keys: Union[str, List[str], None] = None, **kwargs,
    ):
        """Merge on concat dataframes dependending on the keys

        Parameters
        ----------
        keys:
            str or list of str labels of columns to merge on
        """
        super().__init__(**kwargs)

        if keys is None:
            keys = []
        self.keys = maybe_make_list(keys)
        self.complete_df = pd.DataFrame(columns=self.keys)

    def run(self, in_df: List[pd.DataFrame]) -> pd.DataFrame:  # type: ignore
        """
        If values of keys exist on in_df and complete_df will
        merge and add the in_df columns
        else will concat the in_df on the complete_df

        Parameters
        ----------
        in_df
            A pandas DataFrame containing the keys as columns
        Returns
        -------
        A pandas DataFrame
            with merged or concateneted data
        Examples
        --------
        >>> import pandas as pd
        >>> from soam.workflow import MergeConcat
        >>> df1 = pd.DataFrame({"date": [1], "metric1": [512]})
        >>> df2 = pd.DataFrame({"date": [1], "metric2": [328]})
        >>> df3 = pd.DataFrame({"date": [2], "metric1": [238]})
        >>> mc = MergeConcat(keys="date")
        >>> mc.run([df1,df2,df3])
        date	metric1	metric2
        1	    512.0	328.0
        2	    238.0	NaN
        """
        complete_df = pd.DataFrame(columns=self.keys)
        for df in in_df:
            if self._check_keys(df, complete_df):
                if set(df).issubset(set(complete_df.columns)):
                    complete_df = complete_df.combine_first(df)
                else:
                    complete_df = complete_df.merge(df, how="right", on=self.keys)
            else:
                complete_df = pd.concat([complete_df, df])

        return complete_df

    def _check_keys(self, in_df: pd.DataFrame, complete_df: pd.DataFrame) -> bool:
        """
        Check if keys values are in both in_df or complete_df
        """
        df_dict = in_df[self.keys].to_dict("list")
        return any(complete_df.isin(df_dict)[self.keys].all(axis=1))



# console.py
"""
Console
----------
Console commands and utilities. It uses click library to run the app in a
command line interface.
It has a hook inside setup.py:
"entry_points={'console_scripts': ['soam = soam.console:cli']}"
"""
# from sys import argv
from typing import NoReturn

import click
from cookiecutter.main import cookiecutter
from pkg_resources import resource_filename

from soam.cfg import INIT_TEMPLATE_DIR


@click.group()
def cli() -> NoReturn:
    """
    TODO: review why we are using this.
    https://click.palletsprojects.com/en/7.x/commands/#callback-invocation
    """


@cli.command()
@click.option('--output', help='Output directory', default='.')
def init(output: str) -> str:
    """Create a sample project structure ready to use SoaM

    To create the sample project it will prompt the user for the following
     configurations: package_display_name, project_name, package_name,
     author_name, description.

    Parameters
    ----------
    output : str
        The path to make the scaffolding.

    Returns
    -------
    dict
        Returns the root of the sample project: "output/package_name".
    """
    return cookiecutter(
        resource_filename(__name__, INIT_TEMPLATE_DIR), output_dir=output,
    )


# TODO: include __main__ for debugging purposes
# if __name__ == "__main__":
#     assert len(argv) == 2
#     dd = init(['--output', argv[1]])



# plot_utils.py
"""
Plot Utils
----------
Auxiliary functions for the Forecast Plotter.

See Also
--------
forecast_plotter.py.ForecastPlotter : Postprocessor to plot the forecasts.
"""
from copy import deepcopy
from datetime import timedelta
import logging
from typing import Dict, Optional, Tuple

from matplotlib.axes import Axes
import matplotlib.dates as mdates
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
import numpy as np
import pandas as pd

from soam.constants import (
    ANOMALY_WIN,
    ANOMALY_WIN_FILL,
    COLORS,
    DAILY_TIME_GRANULARITY,
    DATE_FORMAT,
    DS_COL,
    FIG_SIZE,
    FONT_SIZE,
    FORECAST,
    HISTORY,
    HISTORY_FILL,
    HOURLY_TIME_GRANULARITY,
    LABELS,
    MAJOR_INTERVAL,
    MAJOR_PAD,
    MINOR_INTERVAL,
    MINOR_LABEL_ROTATION,
    MONTHLY_TIME_GRANULARITY,
    OUTLIER_SIGN_COL,
    OUTLIERS_HISTORY,
    OUTLIERS_NEGATIVE,
    OUTLIERS_POSITIVE,
    PLOT_CONFIG,
    Y_COL,
    YHAT_COL,
    YHAT_LOWER_COL,
    YHAT_UPPER_COL,
)

DAYLOCATOR = "DayLocator"

pd.plotting.register_matplotlib_converters()

logger = logging.getLogger(__name__)


def _set_time_locator_interval(
    fig: Figure, ax: Axes, time_granularity: str, plot_conf: dict
) -> Tuple[Figure, Axes]:
    major_locator = getattr(mdates, DAYLOCATOR)
    major_locator_interval = plot_conf[MAJOR_INTERVAL]
    date_format = "%Y-%b-%d"

    if time_granularity == MONTHLY_TIME_GRANULARITY:
        date_format = "%Y-%b"

    if time_granularity == HOURLY_TIME_GRANULARITY:
        date_format = "%b %a %d"
        minor_locator_interval = plot_conf[MINOR_INTERVAL]
        minor_date_format = "%H"
        ax.xaxis.set_minor_locator(mdates.HourLocator(interval=minor_locator_interval))
        ax.xaxis.set_minor_formatter(mdates.DateFormatter(minor_date_format))

        plt.tick_params(axis="x", which="minor", labelsize=plot_conf[FONT_SIZE])
        ax.xaxis.set_tick_params(
            which="minor", labelrotation=plot_conf[MINOR_LABEL_ROTATION]
        )

        ax.set_xticklabels(ax.get_xticklabels(), fontsize=plot_conf[FONT_SIZE])
        ax.xaxis.set_tick_params(which="major", pad=plot_conf[MAJOR_PAD])

    ax.xaxis.set_major_locator(major_locator(interval=major_locator_interval))
    ax.xaxis.set_major_formatter(mdates.DateFormatter(date_format))

    ax.xaxis_date()  # interpret the x-axis values as dates
    fig.autofmt_xdate()
    return fig, ax


def _base_10_tick_scaler(median_val: float) -> int:
    for i in reversed(range(1, 4)):
        if median_val / (10 ** i) >= 1:
            return i
    return 0


def _create_common_series(
    df: pd.DataFrame, ds_col: str, start_date=None, end_date=None
) -> Tuple[pd.DataFrame, np.ndarray]:  # pylint:disable=unused-argument
    """Get series data for start/end date range.

    Return filtered values and dates."""
    window = df
    if start_date:
        window = window.query(f"@start_date <= {ds_col}")
    if end_date:
        window = window.query(f"{ds_col} <= @end_date")
    window_dates = window[ds_col].dt.to_pydatetime()
    return window, window_dates


def create_forecast_figure(
    df: pd.DataFrame,
    metric_name: str,
    end_date,
    forecast_window,
    anomaly_window: int = 0,
    time_granularity: str = DAILY_TIME_GRANULARITY,
    plot_config: Optional[Dict] = None,
) -> Figure:
    """Plot trend, forecast and anomalies with history, anomaly and forecast phases."""

    plot_config_: Dict
    if plot_config is None:
        plot_config_ = PLOT_CONFIG

    plot_config_ = deepcopy(plot_config)["anomaly_plot"]  # type: ignore

    # plot_config_ = plot_config["anomaly_plot"]
    plot_time_conf: dict = plot_config_[time_granularity]
    color_conf: dict = plot_config_[COLORS]
    label_conf: dict = plot_config_[LABELS]

    anomaly_start_date = end_date - timedelta(days=(anomaly_window))
    future_date = end_date + timedelta(days=forecast_window)

    # Note: we need to convert datetime values in series to pydatetime explicitly
    # See: https://stackoverflow.com/q/29329725/2149400
    history, history_dates = _create_common_series(df, DS_COL, end_date=future_date)
    anomaly_win, anomaly_win_dates = _create_common_series(
        df, DS_COL, start_date=anomaly_start_date, end_date=end_date
    )
    forecast, forecast_dates = _create_common_series(
        df, DS_COL, start_date=end_date, end_date=future_date
    )

    date_format = plot_time_conf[DATE_FORMAT]
    fig_size = plot_time_conf[FIG_SIZE]

    fig, ax = plt.subplots(figsize=fig_size)
    ax.plot(
        history_dates,
        history[Y_COL],
        color=color_conf[HISTORY],
        label=label_conf[HISTORY],
    )
    ax.plot(
        forecast_dates,
        forecast[YHAT_COL],
        ls="--",
        lw=2,
        color=color_conf[FORECAST],
        label=label_conf[FORECAST],
    )

    if YHAT_LOWER_COL in df and YHAT_UPPER_COL in df:
        ax.plot(
            anomaly_win_dates,
            anomaly_win[Y_COL],
            color=color_conf[ANOMALY_WIN],
            label=label_conf[ANOMALY_WIN].format(anomaly_window=anomaly_window),
        )
        ax.fill_between(
            history_dates,
            history[YHAT_LOWER_COL],
            history[YHAT_UPPER_COL],
            color=color_conf[HISTORY_FILL],
            alpha=0.2,
        )

        ax.fill_between(
            anomaly_win_dates,
            anomaly_win[YHAT_LOWER_COL],
            anomaly_win[YHAT_UPPER_COL],
            color=color_conf[ANOMALY_WIN_FILL],
            alpha=0.6,
        )

        ax.fill_between(
            forecast_dates,
            forecast[YHAT_LOWER_COL],
            forecast[YHAT_UPPER_COL],
            color=color_conf[FORECAST],
            alpha=0.2,
        )

        for _, o in df.query(
            f"{DS_COL} >= @history_dates.min() & {OUTLIER_SIGN_COL} != 0"
        ).iterrows():
            o_ds = o[f"{DS_COL}"]
            color = color_conf[OUTLIERS_HISTORY]
            o_label = ""
            if o_ds.to_pydatetime() in np.unique(anomaly_win_dates):
                color = (
                    color_conf[OUTLIERS_POSITIVE]
                    if o[OUTLIER_SIGN_COL] > 0
                    else color_conf[OUTLIERS_NEGATIVE]
                )
                o_label = label_conf["outlier"].format(date=f"{o_ds:{date_format}}")
            ax.plot_date(
                x=o_ds,
                y=o[Y_COL],
                marker="o",
                markersize=6,
                alpha=0.7,
                color=color,
                label=o_label,
            )

    ax.set_xlabel(label_conf["xlabel"])
    ax.set_xlim([history_dates.min(), forecast_dates.max()])

    # FIXME: This seems to break the label ticks.
    # fig, ax = _set_time_locator_interval(fig, ax, time_granularity, plot_time_conf)
    base_10_scale = _base_10_tick_scaler(df[Y_COL].median())
    ax.yaxis.set_major_formatter(
        FuncFormatter(lambda y, pos: f"{(y * (10 ** -base_10_scale)):g}")
    )
    base_10_scale_zeros = base_10_scale * "0"
    ax.set_ylabel(
        label_conf["ylabel"].format(
            metric_name=metric_name, base_10_scale_zeros=base_10_scale_zeros
        )
    )

    if anomaly_window > 0:
        plot_type = "Anomaly"
        start_date = anomaly_start_date
    else:
        plot_type = "Forecast"
        start_date = end_date + timedelta(days=1)
        end_date = future_date
    title = plot_config_["title"].format(
        plot_type=plot_type,
        metric_name=metric_name,
        start_date=start_date,
        end_date=end_date,
    )

    if time_granularity == HOURLY_TIME_GRANULARITY:
        title += f" {end_date:%H}hs"
    ax.set_title(title)
    ax.grid(True, which="major", c=color_conf["axis_grid"], ls="-", lw=1, alpha=0.2)
    ax.legend(loc="upper left", bbox_to_anchor=(1, 1))
    plt.tight_layout()
    return fig



from soam.workflow.backtester import Backtester, compute_metrics
from soam.workflow.forecaster import Forecaster
from soam.workflow.merge_concat import MergeConcat
from soam.workflow.slicer import Slicer
from soam.workflow.time_series_extractor import TimeSeriesExtractor
import soam.workflow.transformer
from soam.workflow.transformer import (
    BaseDataFrameTransformer,
    DummyDataFrameTransformer,
    Transformer,
)



# data_models.py
"""
Data models
----------
Contains the data models for SQLAlchemy used in the DBSaver.

See Also
--------
savers.DBSaver : Saver used to store data and parameters in a database.
"""
from datetime import datetime
import enum

from sqlalchemy import (
    Column,
    DateTime,
    Enum,
    Float,
    ForeignKey,
    Integer,
    Text,
    UniqueConstraint,
)
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy.types as types
from sqlalchemy.types import TIMESTAMP
from sqlalchemy_utils.types.uuid import UUIDType

from soam.cfg import FORECASTER_VALUES_TABLE, SOAM_FLOW_RUN_TABLE, SOAM_TASK_RUNS_TABLE
from soam.constants import TIMESTAMP_COL


def now_getter() -> datetime:
    """Wrapper for datetime.now()
    TODO: check why we need this wrapper.

    Returns
    -------
    datetime
        The current datetime.
    """
    return datetime.now()


def get_default_run_name() -> str:
    """Obtains a run name that contains the current datetime.
    TODO: unused function.

    Returns
    -------
    str
        A concatenated string of 'forecast-run-' and the current datetime
        in isoformat
    """
    return f"forecast-run-{now_getter().isoformat()}"


Base = declarative_base()
# TODO: The rest of this file is not checked to be numpydoc complaint.


class OracleIdentity(types.UserDefinedType):  # pylint:disable=abstract-method
    def get_dbapi_type(self, dbapi):
        return int

    def get_col_spec(self):
        return "NUMBER GENERATED ALWAYS AS IDENTITY ORDER"


class Identity(types.TypeDecorator):  # pylint:disable=abstract-method

    impl = Integer

    def load_dialect_impl(self, dialect):
        if dialect.name == "oracle":
            return dialect.type_descriptor(OracleIdentity)
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect):
        return value

    def process_result_value(self, value, dialect):
        return value


class AbstractIDBase(Base):  # type: ignore # pylint: disable=too-few-public-methods
    """Helper class to add primary keys."""

    __abstract__ = True

    id = Column(Identity, primary_key=True)


class SoamFlowRunSchema(Base):  # type: ignore

    __tablename__ = SOAM_FLOW_RUN_TABLE

    flow_run_id = Column(UUIDType(binary=False), primary_key=True)
    run_date = Column(DateTime, nullable=True)
    start_datetime = Column(DateTime, nullable=True)
    end_datetime = Column(DateTime, nullable=True)


class StepTypeEnum(enum.Enum):
    extract = "extract"
    preprocess = "preprocess"
    forecast = "forecast"
    postprocess = "postprocess"
    custom = "custom"


# Be aware when droping the table you will have to manually drop the type,
# it persist after droping the table:
# DROP TYPE steptypeenum;


class SoamTaskRunSchema(Base):  # type: ignore

    __tablename__ = SOAM_TASK_RUNS_TABLE

    params = Column(Text, nullable=False)
    params_hash = Column(Text, nullable=False)

    task_run_id = Column(UUIDType(binary=False), primary_key=True)
    flow_run_id = Column(
        UUIDType(binary=False),
        ForeignKey(f"{SOAM_FLOW_RUN_TABLE}.flow_run_id"),
        nullable=True,
    )
    step_type = Column(Enum(StepTypeEnum), nullable=False)


class ForecastValues(AbstractIDBase):
    __tablename__ = FORECASTER_VALUES_TABLE
    __table_args__ = (UniqueConstraint("task_run_id", "forecast_date"),)

    task_run_id = Column(
        UUIDType(binary=False),
        ForeignKey(f"{SOAM_TASK_RUNS_TABLE}.task_run_id"),
        nullable=False,
    )
    forecast_date = Column(DateTime, nullable=False)
    yhat = Column(Float, nullable=False)
    yhat_lower = Column(Float)
    yhat_upper = Column(Float)
    y = Column(Float, nullable=True)
    trend = Column(Float)
    outlier_value = Column(Float)
    outlier_sign = Column(Float)


class AbstractTimeSeriesTable(AbstractIDBase):
    """Base table to store time series data.

    This class should be used as base for the table from which timeseries will be
    extracted.
    ```
    class ConcreteTimeSeriesTable(AbstractTimeSeriesTable):
        __tablename__ = "concrete_ts"

        fact_1 = Column(Float)
        dimension_1 = Column()
    ```
    """

    __abstract__ = True
    timestamp = Column(TIMESTAMP_COL, TIMESTAMP, index=True)



"""Module to extract and aggregate time series.

# General TODOs:
- Make quantiles great again:
  https://gitlab.com/mutt_data/tfg-adsplash/-/blob/master/adsplash/store/dataset.py
  They are a nice feature but couldn't get them to work yet.

Notes:
- Dimensional hierarchy can be implemented via snowflake schema [1].
    We could implement this idea via "virtual dimensions" that are really the values we get from
    joining higher in the hierarchy.
- It would be interesting to implement unique counts.

[1] Ralph Kimball, Margy Ross - The Data Warehouse Toolkit (2013)."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

from jinja2 import Template
import pandas as pd

from soam.constants import (
    DONT_AGGREGATE_SYMBOL,
    NEGATE_SYMBOL,
    TIMESTAMP_COL,
    regex_prefix_symbols,
)
from soam.core import Step

if TYPE_CHECKING:
    import datetime as dt
    import muttlib

# Simple column selection templates.
BASE_TEMPLATE = """
    {{ column }} AS {{ alias }}
"""

COMPUTE_SHARE_COLUMN_TEMPLATE = """
    COALESCE(
        {{ constant if constant else 1.0 }} * {{denominator}}::float8 / NULLIF({{numerator}}, 0), 0
    ) AS {{ alias }}
"""

MAX_TEMPLATE = """
    MAX({{ column }})::float8 AS {{ alias }}
"""
SUM_TEMPLATE = """
    SUM({{ column }})::float8 AS {{ alias }}
"""

JOIN_TEMPLATE = """
    JOIN {{table}} ON {{condition}}
"""


class TimeSeriesExtractor(Step):
    db: "muttlib.dbconn.BaseClient"
    table_name: str

    def __init__(
        self,
        db: "muttlib.dbconn.BaseClient",
        table_name: str,
        **kwargs: Dict[str, Any],
    ):
        """Class to handle the dataset retrieval from the PostgreSql database.

        Parameters
        ----------
        db: muttlib.dbconn.BaseClient
            The database connection to use.
        table_name: str
            The table's name.
        """
        super().__init__(**kwargs)

        self.db = db
        self.table_name = table_name

    def extract(self, build_query_kwargs: Dict[str, Any],) -> pd.DataFrame:
        """Extracts aggregated data and return it as a pandas DataFrame.

        Parameters
        ----------
        build_query_kwargs: dict of {str: obj}
            Configuration of the extraction query to be used for the extraction.

        Returns
        -------
        pd.DataFrame
            Extracted data.
        """
        query, kwargs = self.build_query(**build_query_kwargs)
        conn = self.db._connect()  # pylint: disable=protected-access
        df = self.db.to_frame(query, connection=conn, **kwargs)
        if df.empty:
            df = pd.DataFrame(columns=build_query_kwargs["columns"])
        conn.close()
        return df

    # maybe define class type all this arguments?
    def build_query(
        self,
        columns=None,
        prequery: str = "",
        dimensions: List[str] = None,
        dimensions_values: List[str] = None,
        timestamp_col: str = TIMESTAMP_COL,
        start_date: Union["dt.datetime", str] = None,
        end_date: Union["dt.datetime", str] = None,
        order_by: List[str] = None,
        extra_where_conditions: List[str] = None,
        extra_having_conditions: List[str] = None,
        column_mappings: Dict = None,
        aggregated_column_mappings: Dict = None,
        inner_join: Optional[List[Tuple[str, str, str]]] = None,
    ) -> Tuple[str, Dict[str, Any]]:
        """Build the query to extract and aggregated dataset.

        Parameters
        ----------
        columns: list of str
            The columns to retrieve.
        prequery: str
            Query to prepend to the output.
        dimensions: list of str
            The dimensions by which data will be partitioned / aggregated.
            E.g.: ['year', 'month', 'day', 'game', 'type'].
            If a dimension is prefixed with a "!", the dimensions values will
            be negated and the complement rows will be returned.
            E.g.: ['!country'] with ['US'] would return the rows of every
            country except the US.
            If a dimensions is prefixes with a "#", the dimensions values won't
            be used as a group by but as a filter.`
        dimensions_values: str or list of str
            Values to filter the dataset with the indexes matching the
            dimensions indexes.
            If a list is provided for a dimension, all values will be
            considered valid with an OR operator.
            "*" and None won't filter the results.
            However "*" will be recorded as an estimator's / optimizer's
            dimension in the database for record, None won't.
            E.g. for the example dimensions above:
            [2019, 9, None, 'android_flightpilot', 'interstitial']
            would only return september 2019 instertitials data for flight
            pilot android.
            E.g.: doing ['country'], [['BR', 'US']] would return every row
            that is from either BR or US.
        timestamp_col: str
            Name of the column what will hold the timestamp of the fact.
        start_date: datetime
            A start date to filter the rows.
        end_date: datetime
            An inclusive end date to filter the rows.
        order_by:  list of str
            A list of column names to order by.
        extra_where_conditions: list of str
            A list of conditions to be added to the "where" clause.
        extra_having_conditions:  list of str or None
            A list of conditions to be added to the "having" clause.
        column_mappings: TODO: missing doc
        aggregated_column_mappings: dict
            Contains the aggregation functions and aliases to replace the column
            values.
        inner_join: list of tuple of int
            A list of tables to join on, every tuple is expected to contain:
            (table_name, table_alias, complete_condition).
            The table_alias is optional.
            For example:
            [('table_a', 'at', 'initial_table_model.attr_1 = at.attr_1'),
            ('table_b', 'BBB', 'initial_table_model.attr_2 = BBB.222'),
            ('table_c', None, 'initial_table_model.attr_3 = table_c.attr_3')]

        Returns
        -------
        tuple of (str, dict of {str: obj})
            Renderd SQL query to extract data.
        """

        args_maybe_dt = [start_date, end_date]

        for arg in args_maybe_dt:
            arg = pd.to_datetime(arg)

        # Template
        query = """
          {{ prequery }}
          SELECT {{ columns | join(", ") }}
          FROM {{ table_name }}
          {% if join_tables %}
          {% for j_table in join_tables %}
          INNER JOIN {{ j_table.0 }}
          {% if j_table.1 %}
          AS  {{ j_table.1 }}
          {% endif %}
          ON  {{ j_table.2 }}
          {% endfor %}
          {% endif %}
          {% if where %}
          WHERE {{ where | join(" AND ") }}
          {% endif %}
          {% if group_by %}
          GROUP BY {{ group_by | join(", ") }}
          {% endif %}
          {% if having %}
          HAVING {{ having | join(" AND ") }}
          {% endif %}
          {% if order_by %}
          ORDER BY {{ order_by | join(", ") }}
          {% endif %}
        """
        kwargs: Dict[str, Union[int, float, str]] = {}

        if column_mappings is None:
            column_mappings = {}
        if aggregated_column_mappings is None:
            aggregated_column_mappings = {}

        placeholders = {
            "prequery": prequery,
            "columns": "*",
            "table_name": self.table_name,
            "join_tables": inner_join,
            "where": "",
            "group_by": "",
            "having": "",
            "order_by": "",
        }

        # Dimensions negation and check if disable dimension aggregation
        ((dimensions, negate_dimensions_values), (_, dont_aggregate_dimensions)) = (
            self._negate_dimensions(dimensions),
            self._dont_aggregate_dimensions(dimensions),
        )

        # Columns
        col_map = aggregated_column_mappings
        if dimensions is None or all(dont_aggregate_dimensions):
            col_map = column_mappings

        columns = [col_map.get(col, col) for col in columns]

        placeholders["columns"] = columns

        # Where
        where_conds: List[str] = []
        values_conds, values_kwargs = self._filter_dimensions_values(
            dimensions, dimensions_values, negate_dimensions_values,
        )
        kwargs.update(values_kwargs)
        where_conds.extend(values_conds)
        date_conds, date_kwargs = self._filter_date_range(
            start_date, end_date, timestamp_col=timestamp_col
        )
        kwargs.update(date_kwargs)
        where_conds.extend(date_conds)
        if extra_where_conditions:
            where_conds.extend(extra_where_conditions)
        if where_conds:
            placeholders["where"] = where_conds  # type: ignore

        # Having
        having_conds = []
        if extra_having_conditions:
            having_conds.extend(extra_having_conditions)
        if having_conds:
            placeholders["having"] = having_conds  # type: ignore

        # Group by
        if dimensions is not None:
            placeholders["group_by"] = [
                dim  # type: ignore
                for dim, dont in zip(dimensions, dont_aggregate_dimensions)
                if not dont
            ]
        # Order by
        if order_by is not None:
            placeholders["order_by"] = order_by  # type: ignore

        # Render
        sql = Template(query).render(**placeholders)
        kwargs = {
            k: v if not isinstance(v, str) else f"'{v}'" for k, v in kwargs.items()
        }
        # FIXME: Formatting SQL like this is unsafe. We should pass the params to the
        # engine along the query.
        sql = sql % kwargs
        kwargs = {}
        return sql, kwargs

    def dimensions_values(
        self,
        dimensions,
        dimensions_values=None,
        start_date=None,
        end_date=None,
        order_by=None,
    ):
        """Returns the values for the dimensions provided in the dataset.

        Parameters
        ----------
        dimensions:  list of str
            The column names which dimensions values wants to be retrieved.
            E.g.: ['game', 'type'].
        dimensions_values:  str or list of str or None or "*"
            Values to filter the dataset with the indexes matching the
            dimensions indexes.
            If a list is provided for a dimension, all values will be
            considered valid with an OR operator.
            E.g.:
            dimensions=['country', 'ad_network']
            dimensions_values=['us', None]
            Would return all the ('us', ad_network) pairs.
            Or [['br', 'us'], None] would return all pairs of
            (country, ad_network) with the country being one of br or us.
        start_date: datetime
            A start date to filter the rows.
        end_date: datetime
            An inclusive end date to filter the rows.
        order_by:  list of str
            A list of column names to order by.

        Returns
        -------
        list of [tuple of str]
            A list of tuples with the requested columns.
            E.g.: [('android_flightpilot', 'instertitial',
                   ('android_flightpilot', 'rewardedVideo')]
        """
        # Template
        kwargs = {}
        query = """
          SELECT DISTINCT {{ columns | join(", ") }}
          FROM {{ table_name }}
          {% if where %}
          WHERE {{ where | join(" AND ") }}
          {% endif %}
          {% if order_by %}
          ORDER BY {{ order_by | join(", ") }}
          {% endif %}
        """
        placeholders = {
            "columns": dimensions,
            "table_name": self.table_name,
            "where": "",
            "group_by": dimensions,
            "order_by": "",
        }
        # Dimensions negation
        dimensions, negate_dimensions_values = self._negate_dimensions(dimensions)

        # Where
        where_conds, kwargs = self._filter_date_range(start_date, end_date)
        values_conds, values_kwargs = self._filter_dimensions_values(
            dimensions, dimensions_values, negate_dimensions_values,
        )
        kwargs.update(values_kwargs)
        where_conds.extend(values_conds)
        placeholders["where"] = where_conds

        # Order by
        placeholders["order_by"] = order_by

        # Render
        sql = Template(query).render(**placeholders)
        for k, v in kwargs.items():
            if isinstance(v, str):
                kwargs[k] = f"'{v}'"
            elif isinstance(v, tuple):
                if isinstance(v[0], str):
                    v = (f"'{v_i}'" for v_i in v)
                v = ", ".join(v)
                kwargs[k] = f"({v})"

        sql = sql % kwargs
        ret = [list(row) for row in self.db.execute(sql, params=kwargs)]
        return ret

    def _filter_date_range(
        self, start_date=None, end_date=None, timestamp_col=TIMESTAMP_COL,
    ):
        """Returns a list of conditions for a where clause and a dictionary
        with keyword arguments to fill the conditions parameters.

        Parameters
        ----------
        start_date: datetime
            A start date to filter the rows.
        end_date: datetime
            An inclusive end date to filter the rows.

        Returns
        -------
        tuple of (list of str, dict())
             list of str: list of sql conditions for filtering.
            dict: dictionary holding the values of the placeholders in the
                  conditions.
        """
        conds = []
        kwargs = {}
        if start_date is not None:
            conds.append(f"{timestamp_col} >= %(start_date)s")
            kwargs["start_date"] = start_date
            if not isinstance(start_date, str):
                kwargs["start_date"] = start_date.strftime("%Y-%m-%d")
        if end_date is not None:
            conds.append(f"{timestamp_col} <= %(end_date)s")
            kwargs["end_date"] = end_date
            if not isinstance(end_date, str):
                kwargs["end_date"] = end_date.strftime("%Y-%m-%d")
        return conds, kwargs

    def _negate_dimensions(self, dimensions):
        """Returns dimensions without the prefix symbols and a boolean list
        indication if they have to be negated.

        Parameters
        ----------
        dimensions:  list of str or None
            A list of dimension names

        Returns
        -------
        tuple of (list of str, list of bool)
            * List of dimension names with the negatino prefixes "!" removed.
            * List of boolean values indicating if the dimension has to be
              negated or not
        """
        return self._detect_dimensions_prefix_symbol(dimensions, NEGATE_SYMBOL)

    def _dont_aggregate_dimensions(self, dimensions):
        """Returns dimensions without the prefix symbols and a boolean list
        indication if they don't have to be aggregated.

        Parameters
        ----------
        dimensions:  list of str or None
            A list of dimension names

        Returns
        -------
        tuple of (list of str, list of bool)
            * List of dimension names with the negatino prefixes "#" removed.
            * List of boolean values indicating if the dimension should not be
              aggregated.
        """
        return self._detect_dimensions_prefix_symbol(dimensions, DONT_AGGREGATE_SYMBOL)

    def _detect_dimensions_prefix_symbol(
        self, dimensions, prefix_symbol,
    ):
        """Returns dimensions with the prefix symbols removed and a boolean
        list indication if they are prefixed with it.

        Parameters
        ----------
        dimensions:  list of str or None
            A list of dimension names
        prefix_symbol: str
            The prefix symbol (a character) that we want to check if it
            prefixes the dimension name.

        Returns
        -------
        tuple of (list of str, list of bool)
            * List of dimension names with the prefix symbol removed.
            * List of boolean values indicating if the dimension was prefixed
              with the symbol.
        """
        if dimensions:
            has_symbol = []
            # Find symbol
            for dimension in dimensions:
                if dimension:
                    match = regex_prefix_symbols.match(dimension)
                    has_symbol.append(match and prefix_symbol in match[0])
            # Remove prefixes from dimensinos
            dimensions = [
                regex_prefix_symbols.sub("", dimension) for dimension in dimensions
            ]
            return dimensions, has_symbol
        return dimensions, None

    def _filter_dimensions_values(
        self, dimensions, dimensions_values, negate_dimensions_values,
    ):
        """Returns a list of conditions for a where clause and a dictionary
        with keyword arguments to fill the conditions parameters.

        Parameters
        ----------
        dimensions:  list of str or None
            A list of dimension names
        dimensions_values: obj or list of obj or None
            A list with the wanted values for the dimensions.
        negate_dimensions_values: list of bool
            List of boolean values indicating if the dimension has to be
            negated or not

        Returns
        -------
        tuple of (list of str, dict of {str : obj})
             list of str: list of sql conditions for filtering.
            dict: dictionary holding the values of the placeholders in the
                  conditions.
        """
        conds = []
        kwargs = {}
        if dimensions is not None and dimensions_values is not None:
            for name, value, negate in zip(
                dimensions, dimensions_values, negate_dimensions_values
            ):
                # Can we use a standard parsable language instead of this? pyparsing?
                # https://github.com/pyparsing/pyparsing/blob/master/examples/simpleBool.py
                if value is not None and value != "*":
                    operators = ("!=", "NOT IN") if negate else ("=", "IN")
                    if isinstance(value, (list, tuple)):
                        conds.append("%s %s %%(%s)s" % (name, operators[1], name))
                        kwargs[name] = tuple(value)
                    else:
                        conds.append("%s %s %%(%s)s" % (name, operators[0], name))
                        kwargs[name] = value
        return conds, kwargs

    def run(self, build_query_kwargs: Dict[str, Any]) -> pd.DataFrame:  # type: ignore
        return self.extract(build_query_kwargs)



# constants.py
"""
Constants
----------
Constants for the whole project.
"""
from datetime import datetime, timedelta
import re

# Global
PROJECT_NAME = "SoaM"
DS_COL = "ds"
Y_COL = "y"
YHAT_COL = "yhat"
YHAT_LOWER_COL = "yhat_lower"
YHAT_UPPER_COL = "yhat_upper"

PRED_COLS = [DS_COL, YHAT_COL, YHAT_LOWER_COL, YHAT_UPPER_COL, Y_COL]
SEED = 42

FORECAST_DATE = "forecast_date"

SEED = 42

OUTLIER_SIGN_COL = "sign"

# Status
STATUS_ACTIVE = "active"
STATUS_INACTIVE = "inactive"

# Time and Geo Granularity
HOURS = 24
END_HOUR = 23
HOURLY_TIME_GRANULARITY = "H"
DAILY_TIME_GRANULARITY = "D"
MONTHLY_TIME_GRANULARITY = "M"

TIME_GRANULARITY_NAME_MAP = {
    HOURLY_TIME_GRANULARITY: "hourly",
    DAILY_TIME_GRANULARITY: "daily",
}
TIME_GRANULARITIES = list(TIME_GRANULARITY_NAME_MAP.keys())

# Factor levels constants
MAX_FACTOR_LEVELS = 25
MIN_FACTOR_LEVEL_DATAPOINTS = 10

REGRESSOR_PREFIX = "regr_"

# TIME values
YESTERDAY = (datetime.today() - timedelta(days=1)).strftime("%Y-%m-%d")

# Aggregated mail report constants
AGGREGATED_MAIL_IMAGES = ["aggregated_summary"]

# Plot common config keys

HISTORY = "history"
FORECAST = "forecast"
OUTLIER = "outlier"
COLORS = "colors"
ANOMALY_PLOT = "anomaly_plot"
LABELS = "labels"
TITLE = "title"
ANOMALY_WIN = "anomaly_win"
ANOMALY_WIN_FILL = "anomaly_win_fill"
HISTORY_FILL = "history_fill"
OUTLIERS_HISTORY = "outliers_history"
OUTLIERS_POSITIVE = "outliers_positive"
OUTLIERS_NEGATIVE = "outliers_negative"
AXIS_GRID = "axis_grid"


# Daily and Hourly config
FIG_SIZE = "fig_size"
MAJOR_INTERVAL = "major_interval"
MINOR_INTERVAL = "minor_locator_interval"
FUTURE_WINDOW = "future_window"
HISTORY_WINDOW = "history_window"
FONT_SIZE = "font_size"
MINOR_LABEL_ROTATION = "minor_labelrotation"
MAJOR_PAD = "major_pad"
DATE_FORMAT = "date_format"

# Plots config
PLOT_CONFIG = {
    ANOMALY_PLOT: {
        HOURLY_TIME_GRANULARITY: {
            FIG_SIZE: (13, 9),
            MAJOR_INTERVAL: 2,
            FUTURE_WINDOW: 5,
            HISTORY_WINDOW: 14,
            FONT_SIZE: 7,
            MINOR_LABEL_ROTATION: 40,
            MAJOR_PAD: 25,
            MINOR_INTERVAL: 8,
            DATE_FORMAT: "%Y-%b-%d %Hhs",
        },
        DAILY_TIME_GRANULARITY: {
            FIG_SIZE: (10, 6),
            MAJOR_INTERVAL: 3,
            FUTURE_WINDOW: 15,
            HISTORY_WINDOW: 60,
            FONT_SIZE: 7,
            DATE_FORMAT: "%Y-%b-%d",
        },
        MONTHLY_TIME_GRANULARITY: {
            FIG_SIZE: (13, 9),
            MAJOR_INTERVAL: 2,
            FUTURE_WINDOW: 5,
            HISTORY_WINDOW: 14,
            FONT_SIZE: 7,
            MINOR_LABEL_ROTATION: 40,
            MAJOR_PAD: 25,
            MINOR_INTERVAL: 8,
            DATE_FORMAT: "%Y-%b",
        },
        COLORS: {
            HISTORY: "orange",
            HISTORY_FILL: "gray",
            ANOMALY_WIN: "dodgerblue",
            ANOMALY_WIN_FILL: "lightskyblue",
            FORECAST: "blue",
            OUTLIERS_HISTORY: "black",
            OUTLIERS_POSITIVE: "green",
            OUTLIERS_NEGATIVE: "red",
            AXIS_GRID: "gray",
        },
        LABELS: {
            "xlabel": "Date",
            "ylabel": "{metric_name} ({base_10_scale_zeros}s)",
            HISTORY: "History",
            ANOMALY_WIN: "Last {anomaly_window} days",
            FORECAST: "Forecast",
            OUTLIER: "Outlier: {date}",
        },
        "title": "{plot_type} for {metric_name} {start_date:%d-%b} to {end_date:%d-%b}",
    }
}

# Saver Config
FLOW_FILE_NAME = "flow_tasks.csv"
LOCK_NAME = "flow.lock"

# Timeseries Extractor
DONT_AGGREGATE_SYMBOL = "#"
NEGATE_SYMBOL = "!"
PREFIX_SYMBOLS = DONT_AGGREGATE_SYMBOL + NEGATE_SYMBOL  # dims prefix symbols
regex_prefix_symbols = re.compile(f"^[{PREFIX_SYMBOLS }]*")
TIMESTAMP_COL = 'timestamp'



# runner.py
"""
SoamFlow Class
--------------
A class that can execute the pipeline steps and keep track of the whole run
data

TODO: review if SoamFlow is going to be the only class that can run the
 pipeline
"""
from datetime import datetime
from typing import TYPE_CHECKING, Optional  # pylint:disable=unused-import

from prefect import Flow, Task

if TYPE_CHECKING:
    from soam.savers import Saver


class SoamFlow(Flow):
    def __init__(self, saver: "Optional[Saver]" = None, **kwargs):
        super().__init__(**kwargs)
        self.saver = saver
        self.start_datetime: datetime
        self.end_datetime: datetime
        if self.saver is not None:
            self.state_handlers.append(self.saver.save_flow_run)

    def add_task(self, task: Task) -> Task:
        if self.saver is not None:
            if self.saver.save_task_run not in task.state_handlers:
                task.state_handlers.append(self.saver.save_task_run)
        return super().add_task(task)



"""
Forecaster
----------
Forecaster Task that fits a model to series and predicts.
"""
from typing import TYPE_CHECKING, List, Optional, Tuple  # pylint:disable=unused-import

import pandas as pd

from soam.constants import DS_COL, Y_COL
from soam.core import Step

if TYPE_CHECKING:
    from soam.savers import Saver


class Forecaster(Step):
    """Forecaster Task."""

    def __init__(  # type: ignore
        self,
        model,
        savers: "Optional[List[Saver]]" = None,
        output_length: int = 1,
        ds_col: str = DS_COL,
        response_col: str = Y_COL,
        **kwargs,
    ):
        """Wrap a forecasting model to run it inside a pipeline.

        Parameters
        ----------
        model : scikit-learn.base.BaseEstimator
            The model that will be fitted and execute the predictions.
        savers : Optional[List[Saver]], optional
            The saver to store the parameters and state changes, by default None
        output_length : int, optional
            The length of the output to predict, by default 1
        ds_col : str, optional
            The date column name of the input time series DataFrame, by default DS_COL
        response_col : str, optional
            The y column name of the input time series DataFrame, by default Y_COL
        """
        super().__init__(**kwargs)
        if savers is not None:
            for saver in savers:
                self.state_handlers.append(saver.save_forecast)

        self.model = model
        self.output_length = output_length
        self.ds_col = ds_col
        self.response_col = response_col

        self.time_series = pd.DataFrame()
        self.prediction = pd.DataFrame()

    def run(  # type: ignore
        self, time_series: pd.DataFrame,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, object]:
        """
        Execute fit and predict with a given model and time_series DataFrame.

        Parameters
        ----------
        time_series
            A pandas DataFrame containing as minimum the first column
            with DataTime values, the second column the y to predict
            and the other columns more data

        Returns
        -------
        tuple(pandas.DataFrame, pandas.DataFrame, model)
            0 : Predicted Values DataFrame.
            1 : Provided Time Series data (untouched).
            2 : Trained model.
        """
        self.time_series = time_series.copy()

        X, y = self._format_input(time_series)

        X_train = X[: -self.output_length]
        X_pred = X[-self.output_length :]
        self.model.fit(X_train, y)
        self.prediction = self.model.predict(X_pred)
        return self.prediction, self.time_series, self.model

    def _format_input(
        self, time_series: pd.DataFrame
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """Transform time series DataFrame into SoaM compatible format.

        Parameters
        ----------
        time_series : pandas.DataFrame

        Returns
        -------
        pandas.DataFrame
            Formatted DataFrame
        """
        if self.ds_col not in time_series.columns:
            raise ValueError(f"{self.ds_col} not present Time Series columns.")
        if self.response_col not in time_series.columns:
            raise ValueError(f"{self.response_col} not present Time Series columns.")
        time_series = time_series.sort_values(by=self.ds_col)
        return (
            time_series.drop(self.response_col, axis=1),
            time_series[self.response_col],
        )



# helpers.py
# pylint: skip-file
"""Utility functions for the project.

TODO: review this file, it seems unused in the rest of the project.
"""
from contextlib import contextmanager
from datetime import timedelta
import logging
from pathlib import Path

from muttlib.utils import make_dirs, str_to_datetime
import numpy as np
import pandas as pd
from pandas.tseries import offsets
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

from soam.constants import DS_COL, HOURLY_TIME_GRANULARITY, YHAT_COL
from soam.utilities.utils import range_datetime

logger = logging.getLogger(__name__)

RUN_ID_COL = "run_id"


def get_store_dir(base_dir, kpi, prefix, date, end_date=None, sample_size=None):
    """Get parquet cache storage directory."""
    store_dir = base_dir / kpi / prefix
    store_dir /= f'{date:%Y%m%d}{f"-{end_date:%Y%m%d}" if end_date is not None else ""}'
    if sample_size is not None:
        store_dir /= f"sample-{sample_size}"
    return make_dirs(store_dir)


def get_store_file(
    base_dir,
    kpi,
    date,
    sample_size=None,
    end_date=None,
    prefix=None,
    suffix=None,
    ft="pkl",
    extra_dir=None,
    end_date_hour=False,
):
    """Create name to store tabular data query outputs."""
    save_dir = base_dir / kpi
    if extra_dir is not None:
        save_dir /= extra_dir
    save_dir = make_dirs(save_dir)
    outf = f"{date:%Y%m%d}"
    if end_date is not None:
        outf += f"_{end_date:%Y%m%d}"
        if end_date_hour:
            outf += f"T{end_date:%H}"
    if sample_size is not None:
        outf += f"_sample-{sample_size}"
    if prefix is not None:
        outf = f"{prefix}_{outf}"
    if suffix is not None:
        outf = f"{outf}_{suffix}"
    if ft is not None:
        outf += f".{ft}"
    return save_dir / outf


def get_figure_full_path(
    fig_dir,
    target_col,
    fig_name,
    start_date,
    end_date,
    time_granularity,
    granularity,
    suffix=None,
    as_posix=True,
    end_date_hour=False,
):
    """Create figure file-path for given model inputs."""
    save_dir = make_dirs(fig_dir / granularity / time_granularity / fig_name)
    root_name = f"{start_date:%Y%m%d}_{end_date:%Y%m%d}"
    if end_date_hour:
        root_name += f"T{end_date:%H}"
    root_name += f"_{target_col}"
    if suffix is not None:
        root_name += f"_{suffix}"
    fig_path = Path(save_dir, root_name + ".png")
    if as_posix:
        return fig_path.as_posix()
    return fig_path


def create_forecaster_dates(
    end_date, forecaster_train_window, forecaster_future_window
):
    """Process and correct all respective dates for forecaster."""
    # Check not both future and end date as nulls, assert order
    train_w = forecaster_train_window
    future_w = forecaster_future_window
    m = f'Future ("{future_w}") or train ("{train_w}") windows are not geq 0.'
    assert all([future_w > 0, train_w >= 0]), m
    end_date = str_to_datetime(end_date) if isinstance(end_date, str) else end_date
    start_date = end_date - offsets.Day(train_w)
    future_date = end_date + offsets.Day(future_w)
    return start_date, end_date, future_date


def create_anomaly_window_dates(end_date, anomaly_window, time_granularity):
    hourly_offset = True if time_granularity == HOURLY_TIME_GRANULARITY else False
    end_date = str_to_datetime(end_date) if isinstance(end_date, str) else end_date
    anomaly_window_start_date = end_date - timedelta(days=(anomaly_window - 1))
    return list(
        range_datetime(
            anomaly_window_start_date,
            end_date,
            hourly_offset=hourly_offset,
            as_datetime=False,
        )
    )


def insert_df_multiple_clients(
    db_clients, insertion_ids, table_name, df, coseries_ids=None, **kwargs
):
    """Insert a dataframe to a list of database connections."""
    for db_cli in db_clients:
        dialect = db_cli.dialect
        if insertion_ids:
            df["insertion_id"] = insertion_ids[dialect]
        if coseries_ids:
            df["coseries_id"] = coseries_ids[dialect]
        logger.debug(f"Writing to {dialect} {table_name} table...")
        db_cli.insert_from_frame(df, table_name, if_exists="append", **kwargs)
        logger.info(f'Inserted {len(df)} records in {dialect} "{table_name}" table.')
    return


# TODO remove this function when added to muttlib
@contextmanager
def session_scope(engine: Engine, **session_kw):
    """Provide a transactional scope around a series of operations."""

    Session = sessionmaker(bind=engine)
    session = Session(**session_kw)

    try:
        yield session
        session.commit()
    except Exception as err:
        logger.exception(err)
        session.rollback()
        raise
    finally:
        session.close()



import pandas as pd
from pandas._testing import assert_frame_equal

from soam.workflow import MergeConcat


def test_merge_concat():
    df1 = pd.DataFrame({"date": [1], "metric1": [512]})
    df2 = pd.DataFrame({"date": [1], "metric2": [328]})
    df3 = pd.DataFrame({"date": [2], "metric1": [238]})

    dict_result = {
        "date": [1, 2],
        "metric1": [512.0, 238.0],
        "metric2": [328.0, None],
    }
    df_result = pd.DataFrame(dict_result, index=[0, 1])

    mc = MergeConcat(keys="date")
    df_concated = mc.run([df1, df2, df3])

    assert_frame_equal(
        df_result.reset_index(drop=True),
        df_concated.reset_index(drop=True),
        check_dtype=False,
    )



"""Tests for TimeSeriesExtractor.
"""
import os
import unittest
from unittest import main

import pandas as pd
from sqlalchemy import Column
from sqlalchemy.types import Float, Integer, String

from soam.constants import TIMESTAMP_COL
from soam.data_models import AbstractIDBase, AbstractTimeSeriesTable
from soam.workflow import TimeSeriesExtractor
from tests.db_test_case import TEST_DB_CONNSTR, PgTestCase


class ConcreteTimeSeriesTable(AbstractTimeSeriesTable):

    __tablename__ = "test_data"

    # We should really be using the timestamp inherited from AbstractTimeSeriesTable.

    game = Column(String(64))
    country = Column(String(2))
    ad_network = Column(String(64))
    ad_type = Column(String(64))
    placement_id = Column(String(256))
    waterfall_position = Column(Integer())

    total_count_view_opportunity = Column(Integer())
    count_cache_request = Column(Integer())
    count_cache_success = Column(Integer())
    count_view_start = Column(Integer())
    count_view_click = Column(Integer())
    requests = Column(Integer())
    impressions = Column(Integer())
    clicks = Column(Integer())
    revenue = Column(Float())

    # TODO: Add unique constraint for consistency sake.
    #       https://gitlab.com/mutt_data/tfg-adsplash/-/blob/master/adsplash/store/dataset.py#L442


class ConcreteAdNetworkJoinTimeSeriesTable(AbstractIDBase):
    __tablename__ = "test_ad_network_join_data"

    ad_network = Column(String(64))
    ad_network_group = Column(String(64))


class ConcretePlacementIdJoinTimeSeriesTable(AbstractIDBase):
    __tablename__ = "test_placement_id_join_data"

    placement_id = Column(String(256))
    placement_id_group = Column(String(64))


column_mappings = {
    "opportunities": "total_count_view_opportunity AS opportunities",
    "ecpm": """COALESCE(
        1000.0 * revenue::float8 / NULLIF(impressions, 0), 0) AS ecpm""",
    "share": """COALESCE(
        impressions::float8
        / NULLIF(total_count_view_opportunity, 0), 0) AS share""",
    "quantile": """
        COALESCE(1 - (
        SUM(impressions)
        OVER (
            PARTITION BY %(quantile_partition)s
            ORDER BY
            timestamp,
            waterfall_position,
            revenue / NULLIF(impressions, 0)
            DESC
            ROWS BETWEEN unbounded preceding AND current row)
        )::float8 / NULLIF(total_count_view_opportunity, 0), 0) AS quantile
        """,
}
aggregated_column_mappings = {
    "opportunities": """
        MAX(total_count_view_opportunity)::float8
        AS opportunities
        """,
    "total_count_view_opportunity": """
        MAX(total_count_view_opportunity)::float8
        AS total_count_view_opportunity
        """,
    "count_cache_request": """
        SUM(count_cache_request) AS count_cache_request
        """,
    "count_cache_success": """
        SUM(count_cache_success) AS count_cache_success
        """,
    "count_view_start": "SUM(count_view_start) AS count_view_start",
    "count_view_click": "SUM(count_view_click) AS count_view_click",
    "requests": "SUM(requests) AS requests",
    "impressions": "SUM(impressions) AS impressions",
    "clicks": "SUM(clicks) AS clicks",
    "revenue": "SUM(revenue) AS revenue",
    "ecpm": """COALESCE(
        1000.0 * SUM(revenue)::float8
        / NULLIF(SUM(impressions), 0), 0) AS ecpm""",
    "share": """COALESCE(
        SUM(impressions)::float8
        / NULLIF(MAX(total_count_view_opportunity), 0), 0) AS share""",
    "quantile": """
        COALESCE(1 - (
            SUM(SUM(impressions))
            OVER (
            PARTITION BY %(quantile_partition)s
            ORDER BY
                timestamp,
                waterfall_position,
                SUM(revenue) / NULLIF(SUM(impressions), 0) DESC
            ROWS BETWEEN unbounded preceding AND current row)
            )::float8 / NULLIF(MAX(total_count_view_opportunity), 0)::float8,
        0) AS quantile
        """,
}


@unittest.skipIf(not os.getenv(TEST_DB_CONNSTR), f"{TEST_DB_CONNSTR} is not set")
class TestDatasetStore(PgTestCase):
    def _test_load(
        self,
        columns,
        dimensions,
        dimensions_values,
        start_date,
        end_date,
        order_by,
        expected_values,
        extra_where_conditions=None,
        extra_having_conditions=None,
        inner_join=None,
    ):
        df = self.time_series_extractor.extract(
            build_query_kwargs=dict(
                columns=columns,
                dimensions=dimensions,
                dimensions_values=dimensions_values,
                start_date=start_date,
                end_date=end_date,
                order_by=order_by,
                extra_where_conditions=extra_where_conditions,
                extra_having_conditions=extra_having_conditions,
                column_mappings=column_mappings,
                aggregated_column_mappings=aggregated_column_mappings,
                inner_join=inner_join,
            )
        )
        # Fix for backwards compatible with original tests.
        if TIMESTAMP_COL in df.columns and not df.empty:
            df[TIMESTAMP_COL] = df[TIMESTAMP_COL].dt.strftime("%Y-%m-%d")
        self.assertTrue(isinstance(df, pd.DataFrame))
        columns = [c_name.split(".")[-1] for c_name in columns]
        self.assertEqual(df.columns.tolist(), columns)
        self.assertEqual(sorted(df.values.tolist()), sorted(expected_values))

    def test_load_basic_columns_order_by(self):
        columns = [TIMESTAMP_COL, "opportunities", "impressions", "revenue"]
        values = [
            ["2019-09-01", 1000, 100, 20],
            ["2019-09-01", 1000, 200, 30],
            ["2019-09-01", 1000, 300, 40],
            ["2019-09-02", 300, 30, 6],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "impressions"],
            expected_values=values,
        )

    def test_join_load_basic_columns_order_by(self):
        columns = [TIMESTAMP_COL, "opportunities", "tjd.ad_network", "ad_network_group"]
        values = [
            ['2019-09-01', 1000, 'source1', 'source_group_B'],
            ['2019-09-01', 1000, 'source2', 'source_group_A'],
            ['2019-09-01', 1000, 'source2', 'source_group_A'],
            ['2019-09-02', 300, 'source2', 'source_group_A'],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "tjd.ad_network"],
            expected_values=values,
            inner_join=[
                (
                    "test_ad_network_join_data",
                    "tjd",
                    "tjd.ad_network = test_data.ad_network",
                )
            ],
        )

    def test_join_load_basic_columns_order_by_no_alias(self):
        columns = [
            TIMESTAMP_COL,
            "opportunities",
            "test_ad_network_join_data.ad_network",
            "ad_network_group",
        ]
        values = [
            ['2019-09-01', 1000, 'source1', 'source_group_B'],
            ['2019-09-01', 1000, 'source2', 'source_group_A'],
            ['2019-09-01', 1000, 'source2', 'source_group_A'],
            ['2019-09-02', 300, 'source2', 'source_group_A'],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "test_ad_network_join_data.ad_network"],
            expected_values=values,
            inner_join=[
                (
                    "test_ad_network_join_data",
                    None,
                    "test_ad_network_join_data.ad_network = test_data.ad_network",
                )
            ],
        )

    def test_multiple_join_load_basic_columns_order_by(self):
        columns = [
            TIMESTAMP_COL,
            "opportunities",
            "tjd.ad_network",
            "ad_network_group",
            "tpi.placement_id",
            "placement_id_group",
        ]
        values = [
            ['2019-09-01', 1000, 'source1', 'source_group_B', 'z', 'placement_group_1'],
            ['2019-09-01', 1000, 'source2', 'source_group_A', 'b', 'placement_group_2'],
            ['2019-09-01', 1000, 'source2', 'source_group_A', 'a', 'placement_group_1'],
            ['2019-09-02', 300, 'source2', 'source_group_A', 'a', 'placement_group_1'],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "tjd.ad_network"],
            expected_values=values,
            inner_join=[
                (
                    "test_ad_network_join_data",
                    "tjd",
                    "tjd.ad_network = test_data.ad_network",
                ),
                (
                    "test_placement_id_join_data",
                    "tpi",
                    "tpi.placement_id = test_data.placement_id",
                ),
            ],
        )

    def test_join_aggregation_load_basic_columns_no_alias(self):
        columns = ["opportunities", "ad_network_group"]
        values = [[1000.0, 'source_group_A'], [1000.0, 'source_group_B']]
        self._test_load(
            columns=columns,
            dimensions=["ad_network_group"],
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=None,
            expected_values=values,
            inner_join=[
                (
                    "test_ad_network_join_data",
                    None,
                    "test_ad_network_join_data.ad_network = test_data.ad_network",
                )
            ],
        )

    def test_multiple_join_aggregation_basic_columns_order_by(self):
        columns = [
            "opportunities",
            "ad_network_group",
            "placement_id_group",
        ]
        values = [
            [1000.0, 'source_group_B', 'placement_group_1'],
            [1000.0, 'source_group_A', 'placement_group_1'],
            [1000.0, 'source_group_A', 'placement_group_2'],
        ]
        self._test_load(
            columns=columns,
            dimensions=["ad_network_group", "placement_id_group"],
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=None,
            expected_values=values,
            inner_join=[
                (
                    "test_ad_network_join_data",
                    "tjd",
                    "tjd.ad_network = test_data.ad_network",
                ),
                (
                    "test_placement_id_join_data",
                    "tpi",
                    "tpi.placement_id = test_data.placement_id",
                ),
            ],
        )

    def test_multiple_join_aggregation_basic_columns_filtered(self):
        columns = [
            "opportunities",
            "ad_network_group",
            "placement_id_group",
        ]
        values = [
            [1000.0, 'source_group_A', 'placement_group_1'],
            [1000.0, 'source_group_A', 'placement_group_2'],
        ]
        self._test_load(
            columns=columns,
            dimensions=["ad_network_group", "placement_id_group"],
            dimensions_values=['source_group_A', None],
            start_date=None,
            end_date=None,
            order_by=None,
            expected_values=values,
            inner_join=[
                (
                    "test_ad_network_join_data",
                    "tjd",
                    "tjd.ad_network = test_data.ad_network",
                ),
                (
                    "test_placement_id_join_data",
                    "tpi",
                    "tpi.placement_id = test_data.placement_id",
                ),
            ],
        )

    def test_load_basic_columns_aggregation_order_by(self):
        columns = ["opportunities", "impressions", "revenue"]
        values = [
            [1000, 600, 90],
            [300, 30, 6],
        ]
        self._test_load(
            columns=columns,
            dimensions=[TIMESTAMP_COL],
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL],
            expected_values=values,
        )

    # def test_load_composite_columns(self):
    #     columns = [TIMESTAMP_COL, "ecpm", "share", "quantile"]
    #     values = [
    #         ["2019-09-01", 1e3 * 40 / 300, 300 / 1000, 0.40],
    #         ["2019-09-01", 1e3 * 30 / 200, 200 / 1000, 0.70],
    #         ["2019-09-01", 1e3 * 20 / 100, 100 / 1000, 0.90],
    #         ["2019-09-02", 1e3 * 6 / 30, 30 / 300, 0.90],
    #     ]
    #     self._test_load(
    #         columns=columns,
    #         dimensions=None,
    #         dimensions_values=None,
    #         start_date=None,
    #         end_date=None,
    #         order_by=[TIMESTAMP_COL, "ecpm"],
    #         expected_values=values,
    #     )

    # def test_load_composite_columns_aggregation(self):
    #     columns = [TIMESTAMP_COL, "waterfall_position", "ecpm", "share", "quantile"]
    #     values = [
    #         ["2019-09-01", 0, 1e3 * 20 / 100, 100 / 1000, 0.90],
    #         ["2019-09-01", 1, 1e3 * 30 / 200, 200 / 1000, 0.70],
    #         ["2019-09-01", 2, 1e3 * 40 / 300, 300 / 1000, 0.40],
    #         ["2019-09-02", 0, 1e3 * 6 / 30, 30 / 300, 0.90],
    #     ]
    #     self._test_load(
    #         columns=columns,
    #         dimensions=[TIMESTAMP_COL, "waterfall_position"],
    #         dimensions_values=None,
    #         start_date=None,
    #         end_date=None,
    #         order_by=[TIMESTAMP_COL, "waterfall_position"],
    #         expected_values=values,
    #     )

    def test_load_all_columns_aggregation(self):
        columns = [
            TIMESTAMP_COL,
            "game",
            "country",
            "ad_network",
            "ad_type",
            "placement_id",
            "waterfall_position",
            "opportunities",
            "count_cache_request",
            "count_cache_success",
            "count_view_start",
            "count_view_click",
            "requests",
            "impressions",
            "clicks",
            "revenue",
        ]
        dimensions = columns[:7]
        values = [
            [
                "2019-09-01",
                "1",
                "us",
                "source1",
                "video",
                "z",
                2,
                1000.0,
                500,
                310,
                300,
                40,
                500,
                300,
                20,
                40.0,
            ],
            [
                "2019-09-01",
                "1",
                "us",
                "source2",
                "media",
                "b",
                1,
                1000.0,
                200,
                41,
                30,
                3,
                200,
                200,
                40,
                30.0,
            ],
            [
                "2019-09-01",
                "1",
                "us",
                "source2",
                "video",
                "a",
                0,
                1000.0,
                150,
                104,
                100,
                20,
                150,
                100,
                10,
                20.0,
            ],
            [
                "2019-09-02",
                "1",
                "us",
                "source2",
                "video",
                "a",
                0,
                300.0,
                300,
                200,
                34,
                1,
                300,
                30,
                1,
                6.0,
            ],
        ]
        self._test_load(
            columns=columns,
            dimensions=dimensions,
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=dimensions,
            expected_values=values,
        )

    def test_load_from_start_date(self):
        columns = [TIMESTAMP_COL, "opportunities", "impressions", "revenue"]
        values = [
            ["2019-09-02", 300, 30, 6],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date="2019-09-02",
            end_date=None,
            order_by=[TIMESTAMP_COL, "impressions"],
            expected_values=values,
        )

    def test_load_until_end_date(self):
        columns = [TIMESTAMP_COL, "opportunities", "impressions", "revenue"]
        values = [
            ["2019-09-01", 1000, 100, 20],
            ["2019-09-01", 1000, 200, 30],
            ["2019-09-01", 1000, 300, 40],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date=None,
            end_date="2019-09-01",
            order_by=[TIMESTAMP_COL, "impressions"],
            expected_values=values,
        )

    def test_load_date_range(self):
        columns = [TIMESTAMP_COL, "opportunities", "impressions", "revenue"]
        values = [
            ["2019-09-01", 1000, 100, 20],
            ["2019-09-01", 1000, 200, 30],
            ["2019-09-01", 1000, 300, 40],
        ]
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date="2019-09-01",
            end_date="2019-09-01",
            order_by=[TIMESTAMP_COL, "impressions"],
            expected_values=values,
        )

    def test_load_empty(self):
        columns = [TIMESTAMP_COL, "opportunities", "impressions", "revenue"]
        values = []
        self._test_load(
            columns=columns,
            dimensions=None,
            dimensions_values=None,
            start_date="2019-09-03",
            end_date="2019-09-04",
            order_by=[TIMESTAMP_COL, "impressions"],
            expected_values=values,
        )

    def test_load_dimensions_basic_columns(self):
        columns = [TIMESTAMP_COL, "ad_network", "opportunities", "impressions"]
        values = [
            ["2019-09-01", "source1", 1000, 300],
            ["2019-09-01", "source2", 1000, 300],
            ["2019-09-02", "source2", 300, 30],
        ]
        self._test_load(
            columns=columns,
            dimensions=[TIMESTAMP_COL, "ad_network"],
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "ad_network"],
            expected_values=values,
        )

    def test_load_dimensions_values(self):
        columns = [TIMESTAMP_COL, "ad_network", "opportunities", "impressions"]
        values = [
            ["2019-09-01", "source2", 1000, 100],
            ["2019-09-01", "source2", 1000, 200],
            ["2019-09-02", "source2", 300, 30],
        ]
        self._test_load(
            columns=columns,
            dimensions=[TIMESTAMP_COL, "ad_network", "waterfall_position"],
            dimensions_values=[None, "source2", None],
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "impressions"],
            expected_values=values,
        )

    def test_load_multiple_dimensions_values(self):
        columns = [TIMESTAMP_COL, "ad_network", "opportunities", "impressions"]
        values = [
            ["2019-09-01", "source1", 1000, 300],
            ["2019-09-01", "source2", 1000, 100],
            ["2019-09-01", "source2", 1000, 200],
            ["2019-09-02", "source2", 300, 30],
        ]
        self._test_load(
            columns=columns,
            dimensions=[TIMESTAMP_COL, "ad_network", "waterfall_position"],
            dimensions_values=[None, ["source1", "source2"], None],
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "ad_network", "impressions"],
            expected_values=values,
        )

    def test_load_multiple_dimensions_values_with_negation(self):
        columns = [TIMESTAMP_COL, "ad_network", "opportunities", "impressions"]
        values = [
            ["2019-09-01", "source1", 1000, 300],
        ]
        self._test_load(
            columns=columns,
            dimensions=[TIMESTAMP_COL, "ad_network", "!ad_type", "!placement_id"],
            dimensions_values=[None, None, "media", ["a", "x"]],
            start_date=None,
            end_date=None,
            order_by=[TIMESTAMP_COL, "ad_network", "impressions"],
            expected_values=values,
        )

    def test_dimension_values(self):
        ret = self.time_series_extractor.dimensions_values(
            dimensions=["ad_network"],
            dimensions_values=None,
            start_date=None,
            end_date=None,
            order_by=["ad_network"],
        )
        self.assertEqual(ret, [["source1"], ["source2"]])

    def test_dimension_values_filtered(self):
        ret = self.time_series_extractor.dimensions_values(
            dimensions=["ad_network", "waterfall_position"],
            dimensions_values=["source2", None],
            start_date=None,
            end_date=None,
            order_by=["ad_network", "waterfall_position"],
        )
        expected = [
            ["source2", 0],
            ["source2", 1],
        ]
        self.assertEqual(ret, expected)

    def test_dimension_values_multiple_dimensions(self):
        ret = self.time_series_extractor.dimensions_values(
            dimensions=["ad_network", "ad_type"],
            dimensions_values=["*", ["video"]],
            start_date=None,
            end_date=None,
            order_by=["ad_network"],
        )
        expected = [
            ["source1", "video"],
            ["source2", "video"],
        ]
        self.assertEqual(ret, expected)

    def test_builded_query(self):
        columns = [
            "timestamp",
            "game",
            "country",
            "ad_network",
            "ad_type",
            "placement_id",
        ]
        prequery = "SET extra_float_digits = 3;"
        order_by = ["ad_type"]
        start_date = "2019-09-01"
        end_date = "2019-09-02"

        query = self.time_series_extractor.build_query(
            columns=columns,
            prequery=prequery,
            start_date=start_date,
            end_date=end_date,
            order_by=order_by,
        )
        # remove empty spaces and new lines
        returned_query = " ".join(query[0].split())
        return_query = "SET extra_float_digits = 3; SELECT timestamp, game, country, ad_network, ad_type, placement_id FROM test_data WHERE timestamp >= '2019-09-01' AND timestamp <= '2019-09-02' ORDER BY ad_type"
        self.assertEqual(returned_query, return_query)

    @classmethod
    def setUpClass(cls):
        super().setUp(cls)
        cls.time_series_extractor = TimeSeriesExtractor(
            cls.db_client, ConcreteTimeSeriesTable.__tablename__
        )
        engine = cls.db_client.get_engine()
        ConcreteTimeSeriesTable.__table__.create(engine)  # pylint:disable=no-member
        ConcreteAdNetworkJoinTimeSeriesTable.__table__.create(  # pylint:disable=no-member
            engine
        )
        ConcretePlacementIdJoinTimeSeriesTable.__table__.create(  # pylint:disable=no-member
            engine
        )
        query = """
         INSERT INTO test_data
           (timestamp,
            game,
            country,
            ad_network,
            ad_type,
            placement_id,
            waterfall_position,
            total_count_view_opportunity,
            count_cache_request,
            count_cache_success,
            count_view_start,
            count_view_click,
            requests,
            impressions,
            clicks,
            revenue
            )
         VALUES
           ('2019-09-01',
            '1',
            'us',
            'source1',
            'video',
            'z',
            2,
            1000,
            500,
            310,
            300,
            40,
            500,
            300,
            20,
            40
            ),
           ('2019-09-01',
            '1',
            'us',
            'source2',
            'video',
            'a',
            0,
            1000,
            150,
            104,
            100,
            20,
            150,
            100,
            10,
            20
            ),
           ('2019-09-01',
            '1',
            'us',
            'source2',
            'media',
            'b',
            1,
            1000,
            200,
            41,
            30,
            3,
            200,
            200,
            40,
            30
            ),
           ('2019-09-02',
            '1',
            'us',
            'source2',
            'video',
            'a',
            0,
            300,
            300,
            200,
            34,
            1,
            300,
            30,
            1,
            6
            )
        """
        cls.run_query(query)

        query = """
         INSERT INTO test_ad_network_join_data
           (ad_network,
            ad_network_group
            )
         VALUES
           ('source2',
            'source_group_A'
            ),
           ('source1',
            'source_group_B'
            ),
           ('source3',
            'source_group_B'
            )
        """
        cls.run_query(query)

        query = """
         INSERT INTO test_placement_id_join_data
           (placement_id,
            placement_id_group
            )
         VALUES
           ('z',
            'placement_group_1'
            ),
           ('a',
            'placement_group_1'
            ),
           ('b',
            'placement_group_2'
            )
        """
        cls.run_query(query)

    @classmethod
    def tearDownClass(cls):
        # del cls.dataset
        super().tearDown(cls)

    def setUp(self):
        pass

    def tearDown(self):
        pass


if __name__ == "__main__":
    main()



# slack_report.py
"""
Slack Report
----------
Slack reporting and message formatting tools. Its a postprocess that sends the
model forecasts though the slack app.
"""
from asyncio import Future
from pathlib import Path
from typing import Any, Optional, Union

import pandas as pd
import slack
from slack.web.slack_response import SlackResponse

from soam.cfg import get_slack_cred
from soam.constants import DS_COL, YHAT_COL
from soam.core.step import Step

DEFAULT_GREETING_MESSAGE = "Hello everyone! Here are the results of the forecast for the *{metric_name}* metric:\n"
DEFAULT_FAREWELL_MESSAGE = "Cheers!\n SoaM."


class SlackReport:
    def __init__(
        self, channel_id: str, metric_name: str, setting_path: Optional[str],
    ):
        credentials = get_slack_cred(setting_path)
        self.slack_client = slack.WebClient(credentials["token"])
        self.channel_id = channel_id
        self.metric_name = metric_name

    def send_report(
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ) -> Union[Future, SlackResponse]:
        if greeting_message == DEFAULT_GREETING_MESSAGE:
            greeting_message.format(metric_name=self.metric_name)

        summary_entries = []
        summary_entries.append(greeting_message)

        for _, row in prediction.iterrows():
            date = row[DS_COL].strftime('%Y-%b-%d')
            value = "{:.2f}".format(row[YHAT_COL])
            summary_entries.append(f"• *[{date}]* {value}\n")

        summary_entries.append(farewell_message)

        summary_message = "\n".join(summary_entries)  # type: ignore

        return self.slack_client.files_upload(
            channels=self.channel_id,
            file=str(plot_filename),
            initial_comment=summary_message,
            title=f"{self.metric_name} Forecast",
        )


class SlackReportTask(Step, SlackReport):
    def __init__(
        self,
        channel_id: str,
        metric_name: str,
        setting_path: Optional[str],
        **kwargs: Any,
    ):
        Step.__init__(self, **kwargs)  # type: ignore
        SlackReport.__init__(
            self, channel_id, metric_name, setting_path,
        )

    def run(  # type: ignore
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ):
        return self.send_report(
            prediction, plot_filename, greeting_message, farewell_message
        )



"""SOAM setup file"""

import setuptools

import soam

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

#  define 'extra_dependencies'
extra_dependencies = {
    'dev': [
        'flake8-bugbear',
        'flake8-docstrings',
        'bump2version',
        'sphinx',
        'sphinx_rtd_theme',
        'm2r2',
        'pre-commit==2.5.0',
        'isort==4.3.21',
        'black==19.10b0',
        'mypy==0.782',
        'pylint==2.4.4',
        'nox',
    ],
    'test': [
        'interrogate',
        'nox',
        'pytest',
        'pytest-mpl',
        'pytest-xdist',
        'pytest-cov',
        'pytest-html',
        'pytest-mock',
        'hypothesis',
        'psycopg2-binary',
        'nbval',
        'pdftotext==2.1.5',
        'statsmodels<0.12,>=0.11',
    ],
    'slack': ["slackclient",],
    'prophet': ["fbprophet",],
    'pdf_report': ["jupytext==1.10.2", "papermill==2.3.2", "nbconvert==5.6",],
    'gsheets_report': ["muttlib[gsheets]>=1.0,<2"],
}

# create 'all' and 'report' extras
all_extras = []
report_extras = []
for key, extra_dep in extra_dependencies.items():
    if not extra_dep in all_extras:
        all_extras += extra_dep
    if key == 'slack' or 'report' in key and not extra_dep in report_extras:
        report_extras += extra_dep
extra_dependencies.update({'report': report_extras})
extra_dependencies.update({'all': all_extras})

setuptools.setup(
    name="soam",
    version=soam.__version__,
    author="Mutt Data",
    home_page="https://gitlab.com/mutt_data/soam/",
    keywords="anomalies forecasting reporting",
    author_email="info@muttdata.ai",
    description="Tools for time series analysis, plotting and reporting.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    packages=setuptools.find_packages(),
    package_data={"soam": ["resources/*.html", "templates/*", "db_migrations"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    setup_requires=["pytest-runner", "wheel"],
    tests_require=["pytest", "pytest-cov", "pytest-html", "betamax"],
    test_suite='test',
    install_requires=[
        "jinja2",
        "pandas>=1.0.0",
        "Cython<0.29.18,>=0.29",
        "sqlalchemy<1.4.0,>=1.3.0",
        "sqlalchemy_utils",
        "u8darts",
        "alembic",
        "python-decouple",
        "prefect",
        "filelock",
        "click",
        "cookiecutter",
        "muttlib>=1.0,<2",
    ],
    extras_require=extra_dependencies,
    python_requires="~=3.6",
    entry_points={'console_scripts': ['soam = soam.console:cli']},
)
# TODO: check why 'python setup.py develop' is failing to obtain muttlib, but 'pip install -e .' is working



"""
Slicer
----------
A class to create dataframes for aggregations
"""

import logging
from typing import List, Tuple, Union

import pandas as pd
from pandas.core.common import maybe_make_list

from soam.constants import DS_COL
from soam.core import Step

COLUMN = "column"
GROUP = "group"
MODE = [GROUP, COLUMN]


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Slicer(Step):
    def __init__(
        self,
        dimensions: Union[str, List[str], None] = None,
        metrics: Union[str, List[str], None] = None,
        ds_col: str = DS_COL,
        keeps: Union[str, List[str], None] = None,
        **kwargs,
    ):
        """Slice the incoming data upon the given dimensions

        Parameters
        ----------
        dimensions:
            str or list of str labels of categorical columns to slices
        metrics:
            str or list of str labels of metrics columns to slices
        ds_col:
            str of datetime column
        keeps:
            str or list of str labels of columns to keep.
        """
        if dimensions is None:
            dimensions = []
        if metrics is None:
            metrics = []
        if keeps is None:
            keeps = []
        super().__init__(**kwargs)

        self.dimensions = maybe_make_list(dimensions)
        self.metrics = maybe_make_list(metrics)
        self.ds_col = ds_col
        self.keeps = maybe_make_list(keeps)

    def run(self, raw_df: pd.DataFrame) -> List[Tuple[str, pd.DataFrame]]:  # type: ignore
        """
        Slice the given dataframe with the dimensions setted.

        Parameters
        ----------
        raw_df
            A pandas DataFrame containing the raw data to slice
        Returns
        -------
        list[pd.DataFrame]
            DataFrame containing the sliced dataframes.

        Examples
        --------
        >>> df1 = pd.DataFrame(
        >>> {   "date": [1, 2, 1, 2, 3],
        >>>     "country": ["ARG", "ARG", "BRA", "BRA", "BRA"],
        >>>     "color": ["blue", "red", "red", "blue", "blue"],
        >>>     "metric": [451, 213, 378, 754, 546]})
        >>>     "metric2": [333, 444, 555, 666, 777]})
        >>> slicing_test = Slicer(metrics=["metric", "metric2"], ds_col="date", \
                            dimensions=["country","color", ["country","color"]] )
        >>> slicing_test.run(df1)

        [
        date country  metric
          1     ARG     451
          2     ARG     213,
        date country  metric2
          1     ARG      333
          2     ARG      444,
        date country  metric
          1     BRA     378
          2     BRA     754
          3     BRA     546,
        date country  metric2
          1     BRA      555
          2     BRA      666
          3     BRA      777,
        date color  metric
          1  blue     451
          2  blue     754
          3  blue     546,
        date color  metric2
          1  blue      333
          2  blue      666
          3  blue      777,
        date color  metric
          1   red     378
          2   red     213,
        date color  metric2
          1   red      555
          2   red      444,
        date country color  metric
          1     ARG  blue     451,
        date country color  metric2
          1     ARG  blue      333,
        date country color  metric
          2     ARG   red     213,
        date country color  metric2
          2     ARG   red      444,
        date country color  metric
          2     BRA  blue     754
          3     BRA  blue     546,
        date country color  metric2
          2     BRA  blue      666
          3     BRA  blue      777,
        date country color  metric
          1     BRA   red     378,
        date country color  metric2
          1     BRA   red      555]

        """

        dataframes_ret = []

        # Validate logic
        self._check_dimensions(raw_df.columns.tolist())
        if not self.dimensions or not self.metrics:
            raise ValueError("Error no dimension neither metric")

        # Setup dimensinal groups
        raw_df = raw_df.sort_values(by=self.ds_col)

        if self.dimensions:
            groups = []
            for dimension in self.dimensions:
                groups.extend([(g[1], dimension) for g in raw_df.groupby(dimension)])
        else:
            groups = [(raw_df, [])]

        # Make the cuts
        for group, dimension in groups:
            for metric in self.metrics:
                cols = [
                    self.ds_col,
                    *maybe_make_list(dimension),
                    *maybe_make_list(metric),
                    *self.keeps,
                ]
                dataframes_ret.append(group[cols])
        logger.info("Dataframe sliced into %s pieces", len(dataframes_ret))
        return dataframes_ret

    def _check_dimensions(self, columns: List[str]):
        """Check if the dimensions and ds columns are in the dataframe"""
        dimensions = []
        for dim in self.dimensions:
            dimensions.extend(maybe_make_list(dim))
        different_columns = set(dimensions) - set(columns)
        different_columns.update(set([self.ds_col]) - set(columns))

        if len(different_columns) > 0:
            raise ValueError(
                f"""Error unknown columns are setted in dimensions
                or ds_col: {different_columns}"""
            )
